#if !defined(DECFLOAT_H)
#define DECFLOAT_H

#include <deccw.h>

/* These settings are suggested by the draft proposal for
   floating-point arithmetic for 32-bit, 64-bit, and 128-bit storage
   formats.  The 256-bit format is an invention of the mathcw library,
   and temporarily masquerades as a 128-bit type. */

#define DEC_EVAL_METHOD		-1	/* indeterminable */

/* significand precision */
#define DEC32_MANT_DIG		7
#define DEC64_MANT_DIG		16
#define DEC128_MANT_DIG	        34
#define DEC256_MANT_DIG	        DEC128_MANT_DIG

/* Minimum power-of-ten exponents */
#define DEC32_MIN_EXP		-95
#define DEC64_MIN_EXP		-383
#define DEC128_MIN_EXP	        -6143
#define DEC256_MIN_EXP	        DEC128_MIN_EXP

/* Maximum power-of-ten exponents */
#define DEC32_MAX_EXP		96
#define DEC64_MAX_EXP		384
#define DEC128_MAX_EXP	        6144
#define DEC256_MAX_EXP	        DEC128_MAX_EXP

#if defined(HAVE_NATIVE_DECIMAL)

/* Maximum representable finite numbers */
#define DEC32_MAX		9.999999E+96DF
#define DEC64_MAX		9.999999999999999E+384DD
#define DEC128_MAX	        9.999999999999999999999999999999999E+6144DL
#define DEC256_MAX	        DEC128_MAX

/* Difference between 1 and the next larger representable number */
#define DEC32_EPSILON		1E-6DF
#define DEC64_EPSILON		1E-15DD
#define DEC128_EPSILON	        1E-33DL
#define DEC256_EPSILON	        DEC128_EPSILON

/* Minimum normalized positive numbers */
#define DEC32_MIN		1E-95DF
#define DEC64_MIN		1E-383DD
#define DEC128_MIN	        1E-6143DL
#define DEC256_MIN	        DEC128_MIN

/* Minimum denormalized positive numbers */
#define DEC32_DEN		0.000001E-95DF
#define DEC64_DEN		0.000000000000001E-383DD
#define DEC128_DEN	        0.000000000000000000000000000000001E-6143DL
#define DEC256_DEN	        DEC128_DEN

#else  /* fall back to decimal masquerading as binary */

#include <float.h>

/* Maximum representable finite numbers */
#define DEC32_MAX		((decimal_float)FLT_MAX)
#define DEC64_MAX		((decimal_double)DBL_MAX)
#define DEC128_MAX	        ((decimal_long_double)LDBL_MAX)
#define DEC256_MAX	        DEC128_MAX

/* Difference between 1 and the next larger representable number */
#define DEC32_EPSILON		((decimal_float)FLT_EPSILON)
#define DEC64_EPSILON		((decimal_double)DBL_EPSILON)
#define DEC128_EPSILON	        ((decimal_long_double)LDBL_EPSILON)
#define DEC256_EPSILON	        DEC128_EPSILON

/* Minimum normalized positive numbers */
#define DEC32_MIN		((decimal_float)FLT_MIN)
#define DEC64_MIN		((decimal_double)DBL_MIN)
#define DEC128_MIN	        ((decimal_long_double)LDBL_MIN)
#define DEC256_MIN	        DEC128_MIN

/* Minimum denormalized positive numbers */
#define DEC32_DEN		(DEC32_MIN  / (decimal_float)8388608.0)
#define DEC64_DEN		(DEC64_MIN  / (decimal_double)4503599627370496.0)
#define DEC128_DEN	        (DEC128_MIN / (decimal_long_double)9223372036854775808.0L)
				/* divisor for 80-bit format (need
				5192296858534827628530496329220096.0L
				for 128-bit format) */
#define DEC256_DEN	        DEC128_DEN

#endif /* defined(HAVE_NATIVE_DECIMAL) */

#if defined(HAVE_NATIVE_DECIMAL_LONG_LONG_DOUBLE)

/***********************************************************************
  This future 256-bit type using the IBM densely-packed decimal format
  as described in

	http://www2.hursley.ibm.com/decimal/decbits.pdf

  is based on the following encoding:

	IEEE 754 256-bit base-10 floating-point arithmetic
		ECBITS       =         20
		ELIMIT       =    3145727
		Precision    =         70 decimal digits
		EMAX         =    1572864
		EMIN         =   -1572863
		EBIAS        =    1572932
		ENORMALS     =    3145660
		ERANGE       =    3145729
		ESUBNORMALS  =         69
		ETINY        =   -1572932
		EPSILON      = 1e-69
		MINSUBNORMAL = 1e-1572932
		MINNORMAL    = 1e-1572863
		MAXNORMAL    = 9.999999999999999999999999999999999999999999999999999999999999999999999e+1572864

  The DPD format encoding constraint that the number of significand
  digits is exactly 1 + 3*(number of significand bits)/10, together with
  the IEEE 754R design constraints that the number of digits in a
  doubled format size must be at least 2 more, and the exponent range at
  least 8 times larger, than in the original size, make these parameter
  choices the only feasible ones.

  The 256-bit wordsize forces ECBITS to be a multiple of 10, and the
  128-bit decimal long double type has ECBITS = 12.  Increasing ECBITS
  from 20 to 30 results in numeric overflows in the IBM decNumber
  package because it requires an exponent range of 3_221_225_471, beyond
  decNumber's architectural limit of +/-999_999_999.

  A further consideration is that the companion binary type have similar
  precision and range.  At least two encodings have this property:

	IEEE 754 256-bit base-2 floating-point arithmetic
		Precision    =        235 bits
		EBITS        =         21
		EBIAS        =    1048575
		EMIN         =   -1048574
		EMAX         =    1048575
		EPSILON      = 3.622271631530684947003647716355116712273012479255655275833045955974200527437e-71
			     = 0x1p-234
		MINSUBNORMAL = 2.149352521440140678349104359626490660468792953231417698887618724875893292558e-315723
			     = 0x1p-1048808
		MINNORMAL    = 5.933714365125831141761620897483690355217585298398054774069668390898321612637e-315653
			     = 0x1p-1048574
		MAXNORMAL    = 6.741140125499073402269065104704245437620185948532688284694491567674226947983e+315652
			     = 0x1.ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffcp+1048575

	IEEE 754 256-bit base-2 floating-point arithmetic
		Precision    =        234 bits
		EBITS        =         22
		EBIAS        =    2097151
		EMIN         =   -2097150
		EMAX         =    2097151
		EPSILON      = 7.244543263061369894007295432710233424546024958511310551666091911948401054874e-71
			     = 0x1p-233
		MINSUBNORMAL = 6.376821966094394356280978342372544112147758820829395057904582721532147083603e-631376
			     = 0x1p-2097383
		MINNORMAL    = 8.802241541725161332971599031600841190648194283873502074572158835633342928622e-631306
			     = 0x1p-2097150
		MAXNORMAL    = 4.544297019161366309996159590797065043318010399459145627088209557343323734900e+631305
			     = 0x1.ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffcp+2097151

  The decimal format exponent range is 30% to 50% wider than that of the
  corresponding binary format of the same storage size for 32-bit,
  64-bit, and 128-bit sizes, and of somewhat less precision.  If we
  continue this tradition, then the 256-bit binary format MUST have
  T = 235 (the first binary encoding above).

***********************************************************************/

#undef DEC256_DEN
#undef DEC256_EPSILON
#undef DEC256_MANT_DIG
#undef DEC256_MAX
#undef DEC256_MAX_EXP
#undef DEC256_MIN
#undef DEC256_MIN_EXP

#define DEC256_DEN              1e-1572932DLL
#define DEC256_EPSILON          1e-69DLL
#define DEC256_MANT_DIG         70
#define DEC256_MAX              9.999999999999999999999999999999999999999999999999999999999999999999999e+1572864DLL
#define DEC256_MAX_EXP          1572864
#define DEC256_MIN              1e-1572863DLL
#define DEC256_MIN_EXP          -1572863

#endif	/* defined(HAVE_NATIVE_DECIMAL_LONG_LONG_DOUBLE) */

/* Provide better names for the macros to more closely match those in
   <float.h>, and avoid reference to storage sizes in bits */

#define DEC_FLT_DEN		DEC32_DEN
#define DEC_FLT_EPSILON		DEC32_EPSILON
#define DEC_FLT_MANT_DIG	DEC32_MANT_DIG
#define DEC_FLT_MAX		DEC32_MAX
#define DEC_FLT_MAX_EXP		DEC32_MAX_EXP
#define DEC_FLT_MIN		DEC32_MIN
#define DEC_FLT_MIN_EXP		DEC32_MIN_EXP

#define DEC_DBL_DEN		DEC64_DEN
#define DEC_DBL_EPSILON		DEC64_EPSILON
#define DEC_DBL_MANT_DIG	DEC64_MANT_DIG
#define DEC_DBL_MAX		DEC64_MAX
#define DEC_DBL_MAX_EXP		DEC64_MAX_EXP
#define DEC_DBL_MIN		DEC64_MIN
#define DEC_DBL_MIN_EXP		DEC64_MIN_EXP

#define DEC_LDBL_DEN		DEC128_DEN
#define DEC_LDBL_EPSILON	DEC128_EPSILON
#define DEC_LDBL_MANT_DIG	DEC128_MANT_DIG
#define DEC_LDBL_MAX		DEC128_MAX
#define DEC_LDBL_MAX_EXP	DEC128_MAX_EXP
#define DEC_LDBL_MIN		DEC128_MIN
#define DEC_LDBL_MIN_EXP	DEC128_MIN_EXP

#define DEC_LLDBL_DEN		DEC256_DEN
#define DEC_LLDBL_EPSILON	DEC256_EPSILON
#define DEC_LLDBL_MANT_DIG	DEC256_MANT_DIG
#define DEC_LLDBL_MAX		DEC256_MAX
#define DEC_LLDBL_MAX_EXP	DEC256_MAX_EXP
#define DEC_LLDBL_MIN		DEC256_MIN
#define DEC_LLDBL_MIN_EXP	DEC256_MIN_EXP

#endif /* !defined(DECFLOAT_H) */
